@@ -11,6 +11,7 @@
 # XML-based documents.
 #
 # Copyright (c) 2004 by Jordi Vilalta  <jvprat@gmail.com>
+# Copyright (c) 2008 by Nicolas François  <nicolas.francois@centraliens.net>
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -97,10 +98,7 @@ sub shiftline {
             push @textentries, ($after, $ref);
             $line = $before.(shift @textentries);
             $ref .= " ".(shift @textentries);
-            while (@textentries) {
-                my ($r, $l) = (pop @textentries, pop @textentries);
-                $self->unshiftline($l,$r);
-            }
+            $self->unshiftline(@textentries);
         }
     }
 
@@ -271,7 +269,7 @@ to override the default behavior specified by the global "wrap" option.
 Example: WE<lt>chapterE<gt>E<lt>titleE<gt>
 
 Note: This option is deprecated.
-You should use the B<translated> and B<untranslate> options instead.
+You should use the B<translated> and B<untranslated> options instead.
 
 =item attributes
 
@@ -306,19 +304,22 @@ Note: the preprocessor directives must only appear between tags
 
 =item translated
 
-=item unstranslated
-
-Space-separated list of the tags you want to translate or not.
+Space-separated list of the tags you want to translate.
 The tags must be in the form <aaa>, but you can join some (<bbb><aaa>) to
 indicate that the content of the tag <aaa> will only be translated when
-it’s into a <bbb> tag.
+it's into a <bbb> tag.
 
 You can also specify some tag options putting some characters in front of
-the tag hierarchy. For example, you can put ’w’ (wrap) or ’W’ (don’t wrap)
+the tag hierarchy. For example, you can put 'w' (wrap) or 'W' (don't wrap)
 to overide the default behavior specified by the global "wrap" option.
 
 Example: WE<lt>chapterE<gt>E<lt>titleE<gt>
 
+=item untranslated
+
+Space-separated list of the tags you do not want to translate or not. It
+uses the same format as the B<translated> option.
+
 =back
 
 =cut
@@ -340,6 +341,7 @@ sub initialize {
 	$self->{options}{'caseinsensitive'}=0;
 	$self->{options}{'tagsonly'}=0;
 	$self->{options}{'tags'}='';
+	$self->{options}{'break'}='';
 	$self->{options}{'translated'}='';
 	$self->{options}{'untranslated'}='';
 	$self->{options}{'defaulttranslateoption'}='';
@@ -363,6 +365,12 @@ sub initialize {
 			$self->{options}{$opt} = $options{$opt};
 		}
 	}
+	# Default options set by modules. Forbidden for users.
+	$self->{options}{'_default_translated'}='';
+	$self->{options}{'_default_untranslated'}='';
+	$self->{options}{'_default_break'}='';
+	$self->{options}{'_default_inline'}='';
+	$self->{options}{'_default_placeholder'}='';
 
 	#It will maintain the list of the translatable tags
 	$self->{tags}=();
@@ -370,8 +378,11 @@ sub initialize {
 	$self->{untranslated}=();
 	#It will maintain the list of the translatable attributes
 	$self->{attributes}=();
+	#It will maintain the list of the breaking tags
+	$self->{break}=();
 	#It will maintain the list of the inline tags
 	$self->{inline}=();
+	#It will maintain the list of the placeholder tags
 	$self->{placeholder}=();
 	#list of the tags that must not be set in the tags or inline category
 	#by this module or sub-module (unless specified in an option)
@@ -690,8 +701,9 @@ sub tag_trans_doctype {
 
 sub tag_break_close {
 	my ($self,@tag)=@_;
-	my $struct = $self->get_path."<".$self->get_tag_name(@tag).">";
-	if ($self->get_translate_options($struct) =~ m/i/) {
+	my $struct = $self->get_path;
+	my $options = $self->get_translate_options($struct);
+	if ($options =~ m/[ip]/) {
 		return 0;
 	} else {
 		return 1;
@@ -716,7 +728,7 @@ sub tag_trans_close {
 
 sub CDATA_extract {
 	my ($self,$remove)=(shift,shift);
-        my ($eof, @tag) = $self->get_string_until(']]>',{include=>1,unquoted=>1,remove=>$remove});
+        my ($eof, @tag) = $self->get_string_until(']]>',{include=>1,unquoted=>0,remove=>$remove});
 
 	return ($eof, @tag);
 }
@@ -730,7 +742,7 @@ sub CDATA_trans {
 
 sub tag_break_alone {
 	my ($self,@tag)=@_;
-	my $struct = $self->get_path."<".$self->get_tag_name(@tag).">";
+	my $struct = $self->get_path($self->get_tag_name(@tag));
 	if ($self->get_translate_options($struct) =~ m/i/) {
 		return 0;
 	} else {
@@ -751,8 +763,9 @@ sub tag_trans_alone {
 
 sub tag_break_open {
 	my ($self,@tag)=@_;
-	my $struct = $self->get_path."<".$self->get_tag_name(@tag).">";
-	if ($self->get_translate_options($struct) =~ m/i/) {
+	my $struct = $self->get_path($self->get_tag_name(@tag));
+	my $options = $self->get_translate_options($struct);
+	if ($options =~ m/[ip]/) {
 		return 0;
 	} else {
 		return 1;
@@ -786,8 +799,9 @@ in the form E<lt>htmlE<gt>E<lt>bodyE<gt>E<lt>pE<gt>.
 
 sub get_path {
 	my $self = shift;
-	if ( @path > 0 ) {
-		return "<".join("><",@path).">";
+	my @add = @_;
+	if ( @path > 0 or @add > 0 ) {
+		return "<".join("><",@path,@add).">";
 	} else {
 		return "outside any tag (error?)";
 	}
@@ -810,12 +824,12 @@ sub tag_type {
 	if (!defined($line)) { return -1; }
 
 	$self->unshiftline($line,$ref);
+	my ($eof,@lines) = $self->get_string_until(">",{include=>1,unquoted=>1});
+	my $line2 = $self->join_lines(@lines);
 	while (!$found && $i < @tag_types) {
 		($match1,$match2) = ($tag_types[$i]->{beginning},$tag_types[$i]->{end});
 		if ($line =~ /^<\Q$match1\E/) {
 			if (!defined($tag_types[$i]->{f_extract})) {
-				my ($eof,@lines) = $self->get_string_until(">",{include=>1,unquoted=>1});
-				my $line2 = $self->join_lines(@lines);
 #print substr($line2,length($line2)-1-length($match2),1+length($match2))."\n";
 				if (defined($line2) and $line2 =~ /\Q$match2\E>$/) {
 					$found = 1;
@@ -942,39 +956,25 @@ doesn't have options).
 =back
 
 =cut
-
-sub tag_in_list {
-	my ($self,$tag,@list) = @_;
-	my $found = 0;
-	my $i = 0;
-	
-	while (!$found && $i < @list) {
-		my $options;
-		my $element;
-		if ($list[$i] =~ /(.*?)(<.*)/) {
-			$options = $1;
-			$element = $2;
-		} else {
-			$element = $list[$i];
-		}
-		if ($self->{options}{'caseinsensitive'}) {
-			if ( $tag =~ /\Q$element\E$/i ) {
-				$found = 1;
-			}
-		} else {
-			if ( $tag =~ /\Q$element\E$/ ) {
-				$found = 1;
-			}
-		}
-		if ($found) {
-			if ($options) {
-				$found = $options;
-			}
-		} else {
-			$i++;
-		}
+sub tag_in_list ($$$) {
+	my ($self,$path,$list) = @_;
+	if ($self->{options}{'caseinsensitive'}) {
+		$path = lc $path;
 	}
-	return $found;
+
+	while (1) {
+		if (defined $list->{$path}) {
+			if (length $list->{$path}) {
+				return $list->{$path};
+			} else {
+				return 1;
+			}
+		}
+		last unless ($path =~ m/</);
+		$path =~ s/^<.*?>//;
+	} 
+
+	return 0;
 }
 
 =head2 WORKING WITH ATTRIBUTES
@@ -1045,7 +1045,7 @@ sub treat_attributes {
 							$tag[0] = $2;
 						}
 						$complete = 1;
-						if ($self->tag_in_list($self->get_path.$name,@{$self->{attributes}})) {
+						if ($self->tag_in_list($self->get_path.$name,$self->{attributes})) {
 							$text .= $self->found_string($value, $ref, { type=>"attribute", attribute=>$name });
 						} else {
 							print wrap_ref_mod($ref, "po4a::xml", dgettext("po4a", "Content of attribute %s excluded: %s"), $self->get_path.$name, $value)
@@ -1077,15 +1077,24 @@ sub treat_attributes {
 #   w: the content shall be re-wrapped
 #   W: the content shall not be re-wrapped
 #   i: the tag shall be inlined
+#   p: a placeholder shall replace the tag (and its content)
+#
+# A translatable inline tag in an untranslated tag is treated as a translatable breaking tag.
+my %translate_options_cache;
 sub get_translate_options {
 	my $self = shift;
 	my $path = shift;
+
+	if (defined $translate_options_cache{$path}) {
+		return $translate_options_cache{$path};
+	}
+
 	my $options = "";
 	my $translate = 0;
 	my $usedefault = 1;
 
 	my $inlist = 0;
-	my $tag = $self->get_tag_from_list($path, @{$self->{tags}});
+	my $tag = $self->get_tag_from_list($path, $self->{tags});
 	if (defined $tag) {
 		$inlist = 1;
 	}
@@ -1104,16 +1113,10 @@ sub get_translate_options {
 		$translate = 1;
 	}
 
-	$tag = $self->get_tag_from_list($path, @{$self->{inline}});
-	if (defined $tag) {
-		$usedefault = 0;
-		$options .= "i";
-	}
-
 # TODO: a less precise set of tags should not override a more precise one
 	# The tags and tagsonly options are deprecated.
 	# The translated and untranslated options have an higher priority.
-	$tag = $self->get_tag_from_list($path, @{$self->{translated}});
+	$tag = $self->get_tag_from_list($path, $self->{translated});
 	if (defined $tag) {
 		$usedefault = 0;
 		$options = $tag;
@@ -1125,17 +1128,43 @@ sub get_translate_options {
 		$options .= ($self->{options}{'wrap'})?"w":"W";
 	}
 
-	$tag = $self->get_tag_from_list($path, @{$self->{untranslated}});
+	if (not defined $tag) {
+		$tag = $self->get_tag_from_list($path, $self->{untranslated});
+		if (defined $tag) {
+			$usedefault = 0;
+			$options = "";
+			$translate = 0;
+		}
+	}
+
+	$tag = $self->get_tag_from_list($path, $self->{inline});
 	if (defined $tag) {
 		$usedefault = 0;
-		$options = "";
-		$translate = 0;
+		$options .= "i";
+	} else {
+		$tag = $self->get_tag_from_list($path, $self->{placeholder});
+		if (defined $tag) {
+			$usedefault = 0;
+			$options .= "p";
+		}
 	}
 
 	if ($usedefault) {
 		$options = $self->{options}{'defaulttranslateoption'};
 	}
 
+	# A translatable inline tag in an untranslated tag is treated as a
+	# translatable breaking tag.
+	if ($options =~ m/i/) {
+		my $ppath = $path;
+		$ppath =~ s/<[^>]*>$//;
+		my $poptions = $self->get_translate_options ($ppath);
+		if ($poptions eq "") {
+			$options =~ s/i//;
+		}
+	}
+
+	$translate_options_cache{$path} = $options;
 	return $options;
 }
 
@@ -1146,35 +1175,20 @@ sub get_translate_options {
 # The tag (or set of tags) is returned with its options.
 #
 # If no tags could match the path, undef is returned.
-sub get_tag_from_list {
-	my ($self,$path,@list) = @_;
-	my $found = 0;
-	my $i = 0;
-	
-	while (!$found && $i < @list) {
-		my $options;
-		my $element;
-		if ($list[$i] =~ /(.*?)(<.*)/) {
-			$options = $1;
-			$element = $2;
-		} else {
-			$element = $list[$i];
-		}
-		if ($self->{options}{'caseinsensitive'}) {
-			if ( $path =~ /\Q$element\E$/i ) {
-				$found = 1;
-			}
-		} else {
-			if ( $path =~ /\Q$element\E$/ ) {
-				$found = 1;
-			}
-		}
-		if ($found) {
-			return $list[$i];
-		} else {
-			$i++;
-		}
+sub get_tag_from_list ($$$) {
+	my ($self,$path,$list) = @_;
+	if ($self->{options}{'caseinsensitive'}) {
+		$path = lc $path;
 	}
+
+	while (1) {
+		if (defined $list->{$path}) {
+			return $list->{$path}.$path;
+		}
+		last unless ($path =~ m/</);
+		$path =~ s/^<.*?>//;
+	}
+
 	return undef;
 }
 
@@ -1214,9 +1228,7 @@ sub treat_content {
 			if ($tag_types[$type]->{'end'} eq "") {
 				if ($tag_types[$type]->{'beginning'} eq "") {
 					# Opening inline tag
-					my $placeholder_regex = join("|", @{$self->{placeholder}});
-					if (length($placeholder_regex) and
-					    $self->get_tag_name(@tag) =~ m/($placeholder_regex)/) { # FIXME
+					if ($self->get_translate_options($self->get_path($self->get_tag_name(@tag))) =~ m/p/) {
 						# We enter a new holder.
 						# Append a <placeholder#> tag to the current
 						# paragraph, and save the @paragraph in the
@@ -1226,7 +1238,8 @@ sub treat_content {
 						my $sub_translations_ref = $old_holder{'sub_translations'};
 						my @sub_translations = @$sub_translations_ref;
 
-						push @paragraph, ("<placeholder".($#sub_translations+1).">", $text[1]);
+						my $placeholder_str = "<placeholder".($#sub_translations+1).">";
+						push @paragraph, ($placeholder_str, $text[1]);
 						my @saved_paragraph = @paragraph;
 
 						$old_holder{'paragraph'} = \@saved_paragraph;
@@ -1236,9 +1249,12 @@ sub treat_content {
 						my @new_paragraph = ();
 						@sub_translations = ();
 						my %new_holder = ('paragraph' => \@new_paragraph,
+						                  'open' => $text[0],
 						                  'translation' => "",
+						                  'close' => undef,
 						                  'sub_translations' => \@sub_translations);
 						push @save_holders, \%new_holder;
+						@text = ();
 
 						# The current @paragraph
 						# (for the current holder)
@@ -1263,26 +1279,24 @@ sub treat_content {
 						}
 					}
 
-					my $placeholder_regex = join("|", @{$self->{placeholder}});
-					if (length($placeholder_regex) and
-					    $self->get_tag_name(@tag) =~ m/($placeholder_regex)/) {
+					if ($self->get_translate_options($self->get_path($self->get_tag_name(@tag))) =~ m/p/) {
 						# This closes the current holder.
 
-						# We keep the closing tag in the holder paragraph.
-						push @paragraph, @text;
-						@text = ();
-
+						push @path, $self->get_tag_name(@tag);
 						# Now translate this paragraph if needed.
 						# This will call pushline and append the
 						# translation to the current holder's translation.
 						$self->translate_paragraph($translate, @paragraph);
+						pop @path;
 
 						# Now that this holder is closed, we can remove
 						# the holder from the stack.
 						my $holder_ref = pop @save_holders;
 						# We need to keep the translation of this holder
 						my %holder = %$holder_ref;
-						my $translation = $holder{'translation'};
+						$holder{'close'} = $text[0];
+						@text = ();
+						my $translation = $holder{'open'}.$holder{'translation'}.$holder{'close'};
 						# Then we store the translation in the previous
 						# holder's sub_translations array
 						my $old_holder_ref = pop @save_holders;
@@ -1294,7 +1308,6 @@ sub treat_content {
 						# it was before we encountered the holder.
 						my $paragraph_ref = $old_holder{'paragraph'};
 						@paragraph = @$paragraph_ref;
-
 						# restore the holder in the stack
 						$old_holder{'sub_translations'} = \@sub_translations;
 						push @save_holders, \%old_holder;
@@ -1311,27 +1324,6 @@ sub treat_content {
 			# has to be translated
 			push @paragraph, @text;
 		}
-
-		# If the next tag closes the last inline tag, we loop again
-		# (In the case of <foo><bar> being the inline tag, we can't
-		# loop back with the "while" because breaking_tag will check
-		# for <foo><bar><bar>, hence the goto)
-		$type = $self->tag_type;
-		if (    ($tag_types[$type]->{'end'} eq "")
-		    and ($tag_types[$type]->{'beginning'} eq "/") ) {
-			my ($tmpeof, @tag) = $self->extract_tag($type,0);
-			if ($self->get_tag_name(@tag) eq $path[$#path]) {
-				# The next tag closes the last inline tag.
-				# We need to temporarily remove the tag from
-				# the path before calling breaking_tag
-				my $t = pop @path;
-				if (!$tmpeof and !$self->breaking_tag) {
-					push @path, $t;
-					goto NEXT_TAG;
-				}
-				push @path, $t;
-			}
-		}
 	}
 
 	# This strips the extracted strings
@@ -1384,49 +1376,6 @@ sub treat_content {
 	# in the current holder translation.
 	$self->translate_paragraph($translate, @paragraph);
 
-	# Now the paragraph is fully translated.
-	# If we have all the holders' translation, we can replace the
-	# placeholders by their translations.
-	# We must wait to have all the translations because the holders are
-	# numbered.
-	if (scalar @save_holders) {
-		my $holder_ref = pop @save_holders;
-		my %holder = %$holder_ref;
-		my $sub_translations_ref = $holder{'sub_translations'};
-		my $translation = $holder{'translation'};
-		my @sub_translations = @$sub_translations_ref;
-
-		# Count the number of <placeholder\d+> in $translation
-		my $count = 0;
-		my $str = $translation;
-		while ($str =~ m/^.*?<placeholder\d+>(.*)$/s) {
-			$count += 1;
-			$str = $1;
-		}
-
-		if (scalar(@sub_translations) == $count) {
-			# OK, all the holders of the current paragraph are
-			# closed (and translated).
-			# Replace them by their translation.
-			while ($translation =~ m/^(.*?)<placeholder(\d+)>(.*)$/s) {
-				# FIXME: we could also check that
-				#          * the holder exists
-				#          * all the holders are used
-				$translation = $1.$sub_translations[$2].$3;
-			}
-			# We have our translation
-			$holder{'translation'} = $translation;
-			# And there is no need for any holder in it.
-			@sub_translations = ();
-			$holder{'sub_translations'} = \@sub_translations;
-# FIXME: is it alright if a document ends by a placeholder?
-		}
-		# Either we don't have all the holders, either we have the
-		# final translation.
-		# We must keep the current holder at the top of the stack.
-		push @save_holders, \%holder;
-	}
-
 	# Push the trailing blanks
 	if ($blank ne "") {
 		$self->pushline($blank);
@@ -1443,6 +1392,13 @@ sub translate_paragraph {
 	my @paragraph = @_;
 	$translate = $self->get_translate_options($self->get_path);
 
+	while (    (scalar @paragraph)
+	       and ($paragraph[0] =~ m/^\s*\n/s)) {
+		$self->pushline($paragraph[0]);
+		shift @paragraph;
+		shift @paragraph;
+	}
+
 	my $comments;
 	while (@comments) {
 		my ($comment,$eoc);
@@ -1500,6 +1456,54 @@ sub translate_paragraph {
 			$self->pushline($self->recode_skipped_text($para));
 		}
 	}
+	# Now the paragraph is fully translated.
+	# If we have all the holders' translation, we can replace the
+	# placeholders by their translations.
+	# We must wait to have all the translations because the holders are
+	# numbered.
+	if (scalar @save_holders) {
+		my $holder_ref = pop @save_holders;
+		my %holder = %$holder_ref;
+		my $sub_translations_ref = $holder{'sub_translations'};
+		my $translation = $holder{'translation'};
+		my @sub_translations = @$sub_translations_ref;
+
+		# Count the number of <placeholder\d+> in $translation
+		my $count = 0;
+		my $str = $translation;
+		while (    (defined $str)
+		       and ($str =~ m/^.*?<placeholder(\d+)>(.*)$/s)) {
+			$count += 1;
+			$str = $2;
+			if ($sub_translations[$1] =~ m/<placeholder\d+>/s) {
+				$count = -1;
+				last;
+			}
+		}
+
+		if (    (defined $translation)
+		    and (scalar(@sub_translations) == $count)) {
+			# OK, all the holders of the current paragraph are
+			# closed (and translated).
+			# Replace them by their translation.
+			while ($translation =~ m/^(.*?)<placeholder(\d+)>(.*)$/s) {
+				# FIXME: we could also check that
+				#          * the holder exists
+				#          * all the holders are used
+				$translation = $1.$sub_translations[$2].$3;
+			}
+			# We have our translation
+			$holder{'translation'} = $translation;
+			# And there is no need for any holder in it.
+			@sub_translations = ();
+			$holder{'sub_translations'} = \@sub_translations;
+		}
+		# Either we don't have all the holders, either we have the
+		# final translation.
+		# We must keep the current holder at the top of the stack.
+		push @save_holders, \%holder;
+	}
+
 }
 
 
@@ -1520,45 +1524,105 @@ or in the initialize function).
 
 sub treat_options {
 	my $self = shift;
-        
-	$self->{options}{'nodefault'} =~ /\s*(.*)\s*/s;
+
+	if ($self->{options}{'caseinsensitive'}) {
+		$self->{options}{'nodefault'}             = lc $self->{options}{'nodefault'};
+		$self->{options}{'tags'}                  = lc $self->{options}{'tags'};
+		$self->{options}{'break'}                 = lc $self->{options}{'break'};
+		$self->{options}{'_default_break'}        = lc $self->{options}{'_default_break'};
+		$self->{options}{'translated'}            = lc $self->{options}{'translated'};
+		$self->{options}{'_default_translated'}   = lc $self->{options}{'_default_translated'};
+		$self->{options}{'untranslated'}          = lc $self->{options}{'untranslated'};
+		$self->{options}{'_default_untranslated'} = lc $self->{options}{'_default_untranslated'};
+		$self->{options}{'attributes'}            = lc $self->{options}{'attributes'};
+		$self->{options}{'inline'}                = lc $self->{options}{'inline'};
+		$self->{options}{'_default_inline'}       = lc $self->{options}{'_default_inline'};
+		$self->{options}{'placeholder'}           = lc $self->{options}{'placeholder'};
+		$self->{options}{'_default_placeholder'}  = lc $self->{options}{'_default_placeholder'};
+	}
+
+	$self->{options}{'nodefault'} =~ /^\s*(.*)\s*$/s;
 	my %list_nodefault;
 	foreach (split(/\s+/s,$1)) {
-	    $list_nodefault{$_} = 1;
+		$list_nodefault{$_} = 1;
 	}
 	$self->{nodefault} = \%list_nodefault;
 
-	$self->{options}{'tags'} =~ /\s*(.*)\s*/s;
-	my @list_tags;
+	$self->{options}{'tags'} =~ /^\s*(.*)\s*$/s;
+	foreach (split(/\s+/s,$1)) {
+		$_ =~ m/^(.*?)(<.*)$/;
+		$self->{tags}->{$2} = $1 || "";
+	}
+
+	$self->{options}{'break'} =~ /^\s*(.*)\s*$/s;
 	foreach my $tag (split(/\s+/s,$1)) {
-		push @list_tags, $tag
+		$tag =~ m/^(.*?)(<.*)$/;
+		$self->{break}->{$2} = $1 || "";
+	}
+	$self->{options}{'_default_break'} =~ /^\s*(.*)\s*$/s;
+	foreach my $tag (split(/\s+/s,$1)) {
+		$tag =~ m/^(.*?)(<.*)$/;
+		$self->{break}->{$2} = $1 || ""
 			unless $list_nodefault{$tag};
 	}
-	$self->{tags} = \@list_tags;
 
-	$self->{options}{'translated'} =~ /\s*(.*)\s*/s;
-	my @list_translated = split(/\s+/s,$1);
-	$self->{translated} = \@list_translated;
+	$self->{options}{'translated'} =~ /^\s*(.*)\s*$/s;
+	foreach my $tag (split(/\s+/s,$1)) {
+		$tag =~ m/^(.*?)(<.*)$/;
+		$self->{translated}->{$2} = $1 || "";
+	}
+	$self->{options}{'_default_translated'} =~ /^\s*(.*)\s*$/s;
+	foreach my $tag (split(/\s+/s,$1)) {
+		$tag =~ m/^(.*?)(<.*)$/;
+		$self->{translated}->{$2} = $1 || ""
+			unless $list_nodefault{$tag};
+	}
 
-	$self->{options}{'untranslated'} =~ /\s*(.*)\s*/s;
-	my @list_untranslated = split(/\s+/s,$1);
-	$self->{untranslated} = \@list_untranslated;
+	$self->{options}{'untranslated'} =~ /^\s*(.*)\s*$/s;
+	foreach my $tag (split(/\s+/s,$1)) {
+		$tag =~ m/^(.*?)(<.*)$/;
+		$self->{untranslated}->{$2} = $1 || "";
+	}
+	$self->{options}{'_default_untranslated'} =~ /^\s*(.*)\s*$/s;
+	foreach my $tag (split(/\s+/s,$1)) {
+		$tag =~ m/^(.*?)(<.*)$/;
+		$self->{untranslated}->{$2} = $1 || ""
+			unless $list_nodefault{$tag};
+	}
 
-	$self->{options}{'attributes'} =~ /\s*(.*)\s*/s;
-	my @list_attr = split(/\s+/s,$1);
-	$self->{attributes} = \@list_attr;
+	$self->{options}{'attributes'} =~ /^\s*(.*)\s*$/s;
+	foreach my $tag (split(/\s+/s,$1)) {
+		if ($tag =~ m/^(.*?)(<.*)$/) {
+			$self->{attributes}->{$2} = $1 || "";
+		} else {
+			$self->{attributes}->{$tag} = "";
+		}
+	}
 
-	$self->{options}{'inline'} =~ /\s*(.*)\s*/s;
 	my @list_inline;
+	$self->{options}{'inline'} =~ /^\s*(.*)\s*$/s;
 	foreach my $tag (split(/\s+/s,$1)) {
-		push @list_inline, $tag
+		$tag =~ m/^(.*?)(<.*)$/;
+		$self->{inline}->{$2} = $1 || "";
+	}
+	$self->{options}{'_default_inline'} =~ /^\s*(.*)\s*$/s;
+	foreach my $tag (split(/\s+/s,$1)) {
+		$tag =~ m/^(.*?)(<.*)$/;
+		$self->{inline}->{$2} = $1 || ""
 			unless $list_nodefault{$tag};
 	}
-	$self->{inline} = \@list_inline;
 
-	$self->{options}{'placeholder'} =~ /\s*(.*)\s*/s;
-	my @list_placeholder = split(/\s+/s,$1);
-	$self->{placeholder} = \@list_placeholder;
+	$self->{options}{'placeholder'} =~ /^\s*(.*)\s*$/s;
+	foreach my $tag (split(/\s+/s,$1)) {
+		$tag =~ m/^(.*?)(<.*)$/;
+		$self->{placeholder}->{$2} = $1 || "";
+	}
+	$self->{options}{'_default_placeholder'} =~ /^\s*(.*)\s*$/s;
+	foreach my $tag (split(/\s+/s,$1)) {
+		$tag =~ m/^(.*?)(<.*)$/;
+		$self->{placeholder}->{$2} = $1 || ""
+			unless $list_nodefault{$tag};
+	}
 }
 
 =head2 GETTING TEXT FROM THE INPUT DOCUMENT
@@ -1610,11 +1674,11 @@ sub get_string_until {
 		push @text, ($line,$ref);
 		$paragraph .= $line;
 		if ($unquoted) {
-			if ( $paragraph =~ /^((\".*?\")|(\'.*?\')|[^\"\'])*$search.*/s ) {
+			if ( $paragraph =~ /^((\".*?\")|(\'.*?\')|[^\"\'])*$search/s ) {
 				$found = 1;
 			}
 		} else {
-			if ( $paragraph =~ /.*$search.*/s ) {
+			if ( $paragraph =~ /$search/s ) {
 				$found = 1;
 			}
 		}
@@ -1637,7 +1701,7 @@ sub get_string_until {
 			$text[$#text-1] =~ s/\Q$line\E$//s;
 		}
 		if(!$include) {
-			$text[$#text-1] =~ /(.*)($search.*)/s;
+			$text[$#text-1] =~ /^(.*)($search.*)$/s;
 			$text[$#text-1] = $1;
 			$line = $2.$line;
 		}
@@ -1646,11 +1710,7 @@ sub get_string_until {
 		}
 	}
 	if (!$remove) {
-		my $i = $#text;
-		while ($i > 0) {
-			$self->unshiftline ($text[$i-1],$text[$i]);
-			$i -= 2;
-		}
+		$self->unshiftline (@text);
 	}
 
 	#If we get to the end of the file, we return the whole paragraph
@@ -1734,10 +1794,12 @@ L<po4a(7)|po4a.7>, L<Locale::Po4a::TransTractor(3pm)|Locale::Po4a::TransTractor>
 =head1 AUTHORS
 
  Jordi Vilalta <jvprat@gmail.com>
+ Nicolas François <nicolas.francois@centraliens.net>
 
 =head1 COPYRIGHT AND LICENSE
 
-Copyright (c) 2004 by Jordi Vilalta  E<lt>jvprat@gmail.comE<gt>
+ Copyright (c) 2004 by Jordi Vilalta  <jvprat@gmail.com>
+ Copyright (c) 2008 by Nicolas François <nicolas.francois@centraliens.net>
 
 This program is free software; you may redistribute it and/or modify it
 under the terms of GPL (see the COPYING file).
